"""
Tests for ``nitrate.passwords``.
"""

import pytest

from nitrate.passwords import get_required_password, use_in_memory_keyring


def test_gets_existing_password() -> None:
    """
    Getting a password with ``get_required_password`` will succeed if
    the password exists.
    """
    use_in_memory_keyring(initial_passwords={("flickr", "api_key"): "12345"})

    assert get_required_password("flickr", "api_key") == "12345"


def test_throws_if_password_does_not_exist() -> None:
    """
    Trying to get a non-existent password with ``get_required_password``
    will throw a ``RuntimeError``.
    """
    with pytest.raises(RuntimeError, match="Could not retrieve password"):
        get_required_password("doesnotexist", "doesnotexist")



"""
Tests for ``nitrate.comments``.
"""

import pytest

from nitrate.comments import fix_wikipedia_links


class TestFixWikipediaLinks:
    """
    Tests for ``fix_wikipedia_links``.
    """

    @pytest.mark.parametrize(
        "comment_text",
        [
            "",
            "Love old gravestones",
            "Text with\na newline",
            'This is a link to <a href="https://example.com">example.com</a>.',
        ],
    )
    def test_text_is_unchanged(self, comment_text: str) -> None:
        """
        If a comment doesn't contain any Wikipedia links, then it's
        unchanged by the Wikipedia link fixer.
        """
        assert fix_wikipedia_links(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_correct_wikipedia_link_is_unchanged(self) -> None:
        """
        A Wikipedia link that is correct is left unchanged.
        """
        comment_text = (
            '<a href="https://en.wikipedia.org/wiki/Flickr" '
            'rel="noreferrer nofollow">en.wikipedia.org/wiki/Flickr</a>.'
        )

        assert fix_wikipedia_links(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_it_fixes_trailing_period(self) -> None:
        """
        If there's a Wikipedia URL that ends with a period, that period
        is correctly identified as part of the URL.

        See https://github.com/Flickr-Foundation/commons.flickr.org/issues/297
        """
        comment_text = (
            '<a href="https://en.wikipedia.org/wiki/William_Barnes_Jr" '
            'rel="noreferrer nofollow">en.wikipedia.org/wiki/William_Barnes_Jr</a>.'
        )
        expected_text = (
            '<a href="https://en.wikipedia.org/wiki/William_Barnes_Jr." '
            'rel="noreferrer nofollow">en.wikipedia.org/wiki/William_Barnes_Jr.</a>'
        )

        assert fix_wikipedia_links(comment_text) == expected_text

    @pytest.mark.vcr()
    def test_it_skips_unwanted_trailing_period(self) -> None:
        """
        If the trailing period isn't part of the Wikipedia page title,
        that period isn't added to the URL.
        """
        comment_text = 'You’re thinking of <a href="https://en.wikipedia.org/wiki/Longitude">en.wikipedia.org/wiki/Longitude</a>.'

        assert fix_wikipedia_links(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_it_skips_ambiguous_url(self) -> None:
        """
        If there's no Wikipedia page with or without the punctuation,
        it leaves the link as-is -- there's no obviously right thing
        for us to do here.
        """
        comment_text = 'This page <a href="https://en.wikipedia.org/wiki/DoesNotExist" rel="noreferrer nofollow">en.wikipedia.org/wiki/DoesNotExist</a>.'

        assert fix_wikipedia_links(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_it_fixes_disambiguation_parens(self) -> None:
        """
        If there's a Wikipedia URL that has disambiguation parens at
        the end, that extra text is identified as part of the URL.
        """
        # This example comes from Jessamyn's comment here:
        # https://www.flickr.com/photos/twm_news/5257092205/#comment72157720409554465
        #
        # Retrieved 21 January 2025
        comment_text = 'This guy! <a href="https://en.wikipedia.org/wiki/Reg_Dixon_" rel="noreferrer nofollow">en.wikipedia.org/wiki/Reg_Dixon_</a>(comedian)'
        expected_text = 'This guy! <a href="https://en.wikipedia.org/wiki/Reg_Dixon_(comedian)" rel="noreferrer nofollow">en.wikipedia.org/wiki/Reg_Dixon_(comedian)</a>'

        assert fix_wikipedia_links(comment_text) == expected_text



"""
Retrieve passwords from the system keychain.

This provides a thin wrapper around the ``keyring`` module.
"""

import keyring


def get_required_password(service_name: str, username: str) -> str:
    """
    Retrieve a password from the keychain, or throw if it's missing.
    """
    # We wrap this API because keyring will return ``None`` rather
    # than tell you a password is missing, e.g.
    #
    #     >>> import keyring
    #     >>> pw = keyring.get_password("doesnotexist", "doesnotexist")
    #     >>> print(pw)
    #     None
    #
    # It's better to throw an error early than let this empty value
    # propagate into our code and bubble up elsewhere.
    #
    # e.g. if we're calling the Flickr API using an API key retrieved
    # using keyring, we'd rather know immediately that it's empty than
    # get a cryptic "invalid API key" error from the Flickr API.
    #
    password = keyring.get_password(service_name, username)

    if password is None:
        raise RuntimeError(f"Could not retrieve password {(service_name, username)}")

    return password


def use_in_memory_keyring(initial_passwords: dict[tuple[str, str], str]) -> None:
    """
    Create an in-memory keyring with a given set of passwords.

    This is useful for testing, when you want to test the interaction
    between a piece of code and the keychain.
    """

    class InMemoryKeyring(keyring.backend.KeyringBackend):
        """
        A keyring implementation which stores passwords in a dictionary.

        This is for testing only.
        """

        def __init__(self) -> None:
            """
            Create an empty keychain.
            """
            self.passwords: dict[tuple[str, str], str] = {}

        @property
        def priority(self) -> int:  # type: ignore
            """
            This is a required property on ``KeyringBackend`` implementations,
            and is used to decide which backend to use.
            """
            # We set a very high priority, so when this backend is used,
            # it will supersede any others.
            return 1_000_000  # pragma: no cover

        def set_password(self, service_name: str, username: str, password: str) -> None:
            """
            Store a password in the keychain.
            """
            self.passwords[(service_name, username)] = password

        def get_password(self, service_name: str, username: str) -> str | None:
            """
            Retrieve a password from the keychain.
            """
            return self.passwords.get((service_name, username))

        def delete_password(
            self, service_name: str, username: str
        ) -> None:  # pragma: no cover
            """
            Remove a password from the keychain.

            This function isn't currently used as part of the tests, but
            we need it to construct an instance of KeyringBackend.
            """
            del self.passwords[(service_name, username)]

    kr = InMemoryKeyring()

    for (service_name, username), password in initial_passwords.items():
        kr.set_password(service_name, username, password)

    keyring.set_keyring(kr)


__all__ = ["get_required_password"]



"""
Stub file.

You should never be importing ``nitrate`` directly, because it isn't
really a single package -- it's more a collection of different packages
distributed as one library.

You should always import individual files, e.g.

    from nitrate.json import DatetimeDecoder

This file only exists to make Python's package management tools happy,
and to hold the version number.
"""

__version__ = "1.2.0"



"""
Code for dealing with Flickr comments.

In particular, it has some code shared across our
"""

import re
import typing
from xml.etree import ElementTree as ET

import httpx

from .xml import find_required_elem


def fix_wikipedia_links(comment_text: str) -> str:
    """
    Fix Wikipedia links in Flickr comments.

    Flickr's comment parser will try to auto-detect links, and it assumes
    punctuation isn't part of links, but this breaks links to some
    Wikipedia pages, for example:

        https://en.wikipedia.org/wiki/William_Barnes_Jr.

    It will omit the final period from the link, which means it goes to
    the wrong page.

    This function will fix the Wikipedia links auto-detected by Flickr.
    It moves any trailing punctuation that's part of the link inside
    the <a>.  We aren't changing the text of the comment, just the
    auto-detected HTML markup.

    See https://github.com/Flickr-Foundation/commons.flickr.org/issues/297
    """
    for m in re.finditer(
        r'<a href="https://en\.wikipedia\.org/wiki/(?P<slug1>[A-Za-z_]+)"'
        r' rel="noreferrer nofollow">'
        r"en\.wikipedia\.org/wiki/(?P<slug2>[A-Za-z_]+)"
        r"</a>"
        # The suffix can be:
        #
        #   - a full stop, which stands alone
        #   - some text in parentheses, which is a disambiguation string.
        #     This must be some non-empty text in parens.
        #
        r"(?P<suffix>\.|\([^\)]+\))",
        comment_text,
    ):
        print(m)
        # This is a defensive measure, because it was easier than
        # getting lookback groups working in the regex.
        if m.group("slug1") != m.group("slug2"):  # pragma: no cover
            continue

        orig_title = m.group("slug1")

        # If there's a Wikipedia page with this exact title, then the
        # link works and we can leave it as-is.
        if _get_wikipedia_page(orig_title) == "found":
            continue

        # Otherwise, check to see if there's a page with the suffix
        # added -- and if there does, use that as the new link.
        alt_title = orig_title + m.group("suffix")

        print(orig_title)
        print(alt_title)

        if _get_wikipedia_page(alt_title) == "found":
            comment_text = comment_text.replace(
                m.group(0),
                (
                    f'<a href="https://en.wikipedia.org/wiki/{alt_title}" '
                    'rel="noreferrer nofollow">'
                    f"en.wikipedia.org/wiki/{alt_title}</a>"
                ),
            )

    return comment_text


WikipediaPageStatus = typing.Literal["found", "redirected", "not_found"]


def _get_wikipedia_page(title: str) -> WikipediaPageStatus:
    """
    Look up a page on Wikipedia and see whether it:

    1.  Exists, with the given title
    2.  Exists, but the title is normalized/redirected
    3.  Isn't found

    """
    resp = httpx.get(
        "https://en.wikipedia.org/w/api.php",
        params={
            "action": "query",
            "prop": "revisions",
            "titles": title,
            "rvprop": "timestamp",
            "format": "xml",
        },
    )
    resp.raise_for_status()

    # Note: the ElementTree API is not hardened against untrusted XML,
    # but we trust the Wikipedia API enough to try this.
    xml = ET.fromstring(resp.text)

    # The API response will contain a single ``page`` element,
    # like so:
    #
    #    <?xml version="1.0"?>
    #    <api batchcomplete="">
    #      <query>
    #        <normalized>
    #          <n from="William_Barnes_Jr" to="William Barnes Jr"/>
    #        </normalized>
    #        <pages>
    #          <page _idx="-1" ns="0" title="William Barnes Jr" missing=""/>
    #        </pages>
    #      </query>
    #    </api>

    # If the <page> element has the ``missing`` attribute, then there's
    # no such page.
    page = find_required_elem(xml, path=".//page")

    if "missing" in page.attrib:
        return "not_found"

    # If the <page> element has the exact same title as the thing we
    # searched for, then it exists.
    #
    # (Note: Wikipedia replaces spaces with underscores in URLs, which
    # we undo to compare titles.)
    if page.attrib["title"] == title.replace("_", " "):
        return "found"

    # If we found a <page> element but it doesn't have the expected title,
    # we may have been redirected.
    if xml.find(".//normalized") is not None:
        return "redirected"

    # This should never happen in practice so we can't test it, but we
    # include it for the sake of easy debugging if it does.
    else:  # pragma: no cover
        raise RuntimeError(
            f"Unable to parse Wikipedia API response for {title!r}: {resp.text}"
        )



"""
Tests for ``nitrate.xml``.
"""

import xml.etree.ElementTree as ET

import pytest

from nitrate.xml import find_required_elem, find_required_text, find_optional_text


XML = ET.fromstring(
    """
    <?xml version="1.0" encoding="utf-8" ?>
    <greeting>
        <english>Hello world</english>
        <french/>
        <hungarian></hungarian>
    </greeting>""".strip()
)


def test_find_required_elem() -> None:
    """
    If we look for a required element that exists, we find it.
    """
    english = find_required_elem(XML, path=".//english")
    assert english is not None
    assert english.text == "Hello world"


def test_find_required_elem_throws_if_cannot_find_element() -> None:
    """
    Looking for a required element that doesn't exist
    throws a ``ValueError``.
    """
    with pytest.raises(ValueError, match="Could not find required match"):
        find_required_elem(XML, path=".//german")


def test_find_required_text() -> None:
    """
    If we look for text in a required element that exists, we find it.
    """
    assert find_required_text(XML, path=".//english") == "Hello world"


def test_find_required_text_throws_if_finds_element_without_text() -> None:
    """
    Looking for text in a required element that exists but is empty
    throws a ``ValueError``.
    """
    with pytest.raises(ValueError, match="Could not find required text"):
        find_required_text(XML, path=".//french")


def test_find_required_text_throws_if_cannot_find_element() -> None:
    """
    Looking for text in a required element that doesn't exist
    throws a ``ValueError``.
    """
    with pytest.raises(ValueError, match="Could not find required match"):
        find_required_text(XML, path=".//german")


@pytest.mark.parametrize(
    ["path", "expected"],
    [
        ("english", "Hello world"),
        ("french", None),
        ("german", None),
        ("hungarian", None),
    ],
)
def test_find_optional_text(path: str, expected: str | None) -> None:
    """
    Looking for optional text will find the text if:

    1.  The element exists, and
    2.  It's non-empty

    """
    assert find_optional_text(XML, path=path) == expected



"""
Some helper methods for writing type-safe Python.
"""

import functools
import json
import pathlib
import typing

from pydantic import ConfigDict, TypeAdapter


T = typing.TypeVar("T")


@functools.cache
def _get_validator(model: type[T]) -> TypeAdapter[T]:
    """
    Get the validator for a given type.  This is a moderately expensive
    process, so we cache the result -- we only need to create the
    validator once for each type.
    """
    try:
        model.__pydantic_config__ = ConfigDict(extra="forbid")  # type: ignore
    except AttributeError:
        pass

    return TypeAdapter(model)


def validate_type(t: typing.Any, *, model: type[T]) -> T:
    """
    Check that some data matches a given type.

    We use this to e.g. check that the structured data we receive from
    Wikimedia matches our definitions, so we can use the data in our
    type-checked Python.

    See https://stackoverflow.com/a/77386216/1558022
    """
    # This is to fix an issue from the type checker:
    #
    #     Argument 1 to "__call__" of "_lru_cache_wrapper"
    #     has incompatible type "type[T]"; expected "Hashable"
    #
    assert isinstance(model, typing.Hashable)

    validator = _get_validator(model)

    return validator.validate_python(t, strict=True)


def read_typed_json(
    path: pathlib.Path | str,
    *,
    model: type[T],
    cls: type[json.JSONDecoder] | None = None,
) -> T:
    """
    Read a JSON file and validate that its contents contain the
    correct type.
    """
    with open(path) as in_file:
        t = json.load(in_file, cls=cls)

    return validate_type(t, model=model)


__all__ = ["validate_type", "read_typed_json"]
